using System;
using System.ServiceProcess;															// Windows Service
using System.Net;																					// IPEndPoint, IPAddres, etc...
using System.Net.Sockets;																	// Sockets
using System.Xml;																					// XmlDocument, XmlNode, etc...
using System.Threading;																		// Threads

namespace UPSLT
{
	namespace Services
	{
		namespace Chat
		{
			public class UPSLTChatServer : System.ServiceProcess.ServiceBase
			{
				// private member variables
				private Int16 m_nPort;															// Port the service will run on
				private int m_nBackLog, m_nBufferSize;							// BackLog = # of connections allowed to wait in queue
																														// BufferSize = size (in bytes) of data sent to/from client & server
				private ManualResetEvent m_allDone;									// Controls the listener thread
				private ChatClientCollection m_clientCollection;		// Colelction of connected clients
			
				public UPSLTChatServer()
				{
					InitializeComponent();
					
					// Initialize member variables (the following values are default values if no command line args are specified)
					m_nPort = 2000;
					m_nBackLog = 10;
					m_nBufferSize = 32;
					m_allDone = new ManualResetEvent(true);
					m_clientCollection = new ChatClientCollection();
				}

				// The main entry point for the process
				static void Main()
				{
					System.ServiceProcess.ServiceBase[] ServicesToRun;		
					ServicesToRun = new System.ServiceProcess.ServiceBase[] { new UPSLTChatServer() };
					System.ServiceProcess.ServiceBase.Run(ServicesToRun);
				}

				
				// VS.NET Generated Method - Required method for Designer support.  Do not modify.
				private void InitializeComponent()
				{
					this.ServiceName = "UPSLT Chat Server";
				}


				protected override void Dispose(bool bDisposing)
				{
					base.Dispose (bDisposing);
				}


				// Start the service.
				protected override void OnStart(string[] args)
				{
					LoadStartupParameters(args);
					Thread t = new Thread(new ThreadStart(ListenForClients_Proc));
					t.Start();
				}	 
				
				
				// Stop the service.			
				protected override void OnStop()
				{
					// TODO: Add code here to perform any tear-down necessary to stop your service.
				}
				
				
				//--------------------------
				// Private Member Functions
				//--------------------------
				
				// load startup parameters			
				private void LoadStartupParameters(string[] _args)
				{
					for (int i = 0; i < _args.Length; i += 2)
					{
						switch (_args[i].ToLower())
						{
							case "-port":
								m_nPort = Int16.Parse(_args[i+1]);
								break;
							case "-backlog":
								m_nBackLog = Int32.Parse(_args[i+1]);
								break;
							case "-buffersize":
								m_nBufferSize = Int32.Parse(_args[i+1]);
								break;
						}
					}		
				}
				
				
				// listen for incoming client connections
				private void ListenForClients_Proc()
				{
					// This thread procedure just waits for client connections,
					// and when it receives one, hands it off, and continues waiting for more connections
				
					// Create a listener socket and a local end point
					Socket listener = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
					IPEndPoint epLocal = new IPEndPoint(IPAddress.Any, m_nPort);
				
					try
					{				
						listener.Bind(epLocal);
						listener.Listen(m_nBackLog);
					
						while (true)
						{
							// Block until a connection is received,
							// Then deal with the connectio in OnClient Connect
							m_allDone.Reset();				
							listener.BeginAccept(new AsyncCallback(OnClientConnect), listener);					
							m_allDone.WaitOne();
						}
					}
					catch(Exception e)
					{
						Log(e.ToString());
					}				
				}
				
				// this function is called when a client connection is detected
				private void OnClientConnect(IAsyncResult async)
				{
					// Pull a copy of the listener socket from the AsyncResult,
					// Get a copy of the socket that just connected to the listener			
					Socket listener = (Socket)async.AsyncState;						
					Socket worker = listener.EndAccept(async);
				
					// Tell the main thread to continue listening for client connections
					m_allDone.Set();
				
					// Create a new chat client based on the worker socket,
					// Listen for incoming data from that client			
					ChatClient client = new ChatClient(worker);
					ListenForData(client);
				}
				
				private void ListenForData(ChatClient _client)
				{
					// Reset the client's data buffer, and begin receiving data into the buffer
					byte[] buffer = _client.ResetByteBuffer(m_nBufferSize);
				
					try
					{
						_client.Pipe.BeginReceive(buffer, 0, buffer.Length, SocketFlags.None, new AsyncCallback(OnDataReceived), _client);
					}
					catch (SocketException se)
					{
						Log(se.Message);
					}
					catch (ObjectDisposedException ode)
					{
						Log(ode.Message);
					}
				}
				
				// this function is executed when the server receives data from a client
				private void OnDataReceived(IAsyncResult async)
				{
					// Get the client (and the client's socket) that sent the data
					ChatClient client = (ChatClient)async.AsyncState;
					Socket clientPipe = client.Pipe;
				
					// Get the number of bytes sent,
					// Convert the byte data to a string
					// Append the newly received data to the client's StringBuffer
					int bytes = clientPipe.EndReceive(async);
					string strBuffer = System.Text.Encoding.ASCII.GetString(client.ByteBuffer, 0, bytes);
					client.AppendStringBuffer(strBuffer);
				
					// If there is a null character, '\0' in the StringBuffer, it means we've received
					// a complete message from the client, so we'll go ahead and process that data now
					if (strBuffer.IndexOf('\0') != -1)
					{
						ProcessData(client);
						client.ResetStringBuffer();					// reset the StringBuffer when we're done processing data
					}
				
					// If we received any bytes, the client is (probably) still connected, so continues
					// listening for incoming data
					if (bytes > 0)
						ListenForData(client);
					else
					{
						// If we received 0 byte data, the client connection (probably) dropped, so Disconnect
						// the client, and remove it from the ChatClientCollection (if it was logged in)
						Logout(client);	
					}
				}
				
				// handle the data that was sent from the client
				private void ProcessData(ChatClient _client)
				{
					// Two important things to note:
					// 1. If the client is not logged in, any commands sent to the server will be rejected
					// 2. The obvious exception to this is the <login...> command
					
					// Data sent from the client is stored in the StringBuffer property of the client object
					string strXml = _client.StringBuffer;
					
					// ProcessData() is only called when a complete message is received (i.e.  the message contains
					// a null, '\0' character.  Therefore, if the message's length is only 1, we know that all we've
					// received is a null character.  Since there is no point is processing a null character, we
					// simply return without doing anything
					if (strXml.Length == 1)
						return;
						
					// If the message's length is greater, we assume that there is some valid message/command to process
					XmlDocument doc = new XmlDocument();
					try
					{
						doc.LoadXml(strXml);
						XmlNode root = doc.DocumentElement;
						string cmd = root.Name.ToLower();
						
						if (!_client.LoggedIn && cmd != "login")
							SendReply(_client, ReplyCode.UNLI);
						else
						{
							switch (cmd)
							{
								case "login":								
									Login(_client, root);
									break;
									
								case "users":
									SendUserList(_client);
									break;	
								
								case "msg":
									Message(_client, root);
									break;
																
								default:
									SendReply(_client, ReplyCode.UNKC);
									break;
							}
						}
					}
					catch (XmlException)
					{
						SendReply(_client, ReplyCode.INVC);
					}				
				}
				
				// Send a user message to specified clients
				private void Message(ChatClient _from, XmlNode _xml)
				{					
					string from = _from.Nickname;										
					string msg = _xml.InnerXml;
					string strXml = "<usrmsg from = \"" + from + "\">" +  msg + "</usrmsg>";
					byte[] data = GetByteData(strXml);
					SendRawData(m_clientCollection, data);						
				}
				
				// Send the list of connected users to the entire chatroom 
				private void SendUserList()
				{
					string strXml = "<users>";
					foreach (ChatClient client in m_clientCollection)
						strXml += "<nickname>" + client.Nickname + "</nickname>";
						
					strXml += "</users>";
					
					byte[] data = GetByteData(strXml);
					SendRawData(m_clientCollection, data);
				}
				
				// Send the list of connected users to a specified client
				private void SendUserList(ChatClient _client)
				{
					string strXml = "<users>";
					foreach (ChatClient client in m_clientCollection)
						strXml += "<nickname>" + client.Nickname + "</nickname>";
						
					strXml += "</users>";
					
					byte[] data = GetByteData(strXml);
					SendRawData(_client, data);
				}
				
				// Try to login a specified client
				private void Login(ChatClient _client, XmlNode _credentials)
				{
					string strNickname = "";
					
					try
					{
						strNickname = _credentials.Attributes["nick"].Value;						
					}
					catch (NullReferenceException)
					{
						SendReply(_client, ReplyCode.LFBA);
						return;
					}
					
					if (m_clientCollection.Contains(_client))
						SendReply(_client, ReplyCode.LFAL);
					else if (m_clientCollection.NickTaken(strNickname))
						SendReply(_client, ReplyCode.LFNT);
					else
					{
						_client.Nickname = strNickname;
						_client.LoggedIn = true;
						m_clientCollection.Add(_client);
						SendReply(_client, ReplyCode.LS);
						
						string strXmlData = "<Nickname>" + strNickname + "</Nickname>";
						BroadCast(MessageCode.ULI, strXmlData);
						SendUserList();
					}
				}
				
				// Logout a specified client
				private void Logout(ChatClient _client)
				{					
					string strXmlData = "<Nickname>" + _client.Nickname + "</Nickname>";
					m_clientCollection.Remove(_client);
					BroadCast(MessageCode.ULO, strXmlData);
					SendUserList();
				}
				
				// Broadcast a message from the server to all connected clients
				private void BroadCast(MessageCode _mc, string _xmlData)
				{
					string strMessage = "";
					
					switch(_mc)
					{
						case MessageCode.ULI:
							strMessage = "<svrmsg code = \"uli\">" + _xmlData + "</svrmsg>";
							break;
						
						case MessageCode.ULO:
							strMessage = "<svrmsg code = \"ulo\">" + _xmlData + "</svrmsg>";
							break;
					}
					
					byte[] data = GetByteData(strMessage);
					SendRawData(m_clientCollection, data);
				}
				
				// Send a reply to a specified client
				private void SendReply(ChatClient _client, ReplyCode _rc)
				{
					string strReply = "";
					
					switch (_rc)
					{
						case ReplyCode.INVC:
							strReply = "<reply code = \"invc\" />";
							break;
							
						case ReplyCode.UNKC:						
							strReply = "<reply code = \"unkc\" />";
							break;
							
						case ReplyCode.UNLI:
							strReply = "<reply code = \"unli\" />";
							break;
							
						case ReplyCode.LFBA:
							strReply = "<reply code = \"lfba\" />";
							break;
							
						case ReplyCode.LFAL:
							strReply = "<reply code = \"lfal\" />";
							break;
							
						case ReplyCode.LFNT:
							strReply = "<reply code = \"lfnt\" />";
							break;
							
						case ReplyCode.LS:
							strReply = "<reply code = \"ls\" />";
							break;
					}
					
					byte[] data = GetByteData(strReply);
					SendRawData(_client, data);
				}
				
				// Get byte data suitable for sending across a TCP Socket
				private byte[] GetByteData(string _str)
				{
					_str += "\0";
					byte[] data = System.Text.Encoding.ASCII.GetBytes(_str);
					return data;
				}
				
				// send raw byte data to a specified client
				private void SendRawData(ChatClient _client, byte[] _data)
				{					
					try
					{
						_client.Pipe.Send(_data);
					}
					catch (ObjectDisposedException)
					{
						// this particular client no longer holds an active connection, so remove him from the collection
						Logout(_client);
					}
					catch (SocketException)
					{
						// ???						
					}
				}
				
				// send raw byte data to a collection of clients
				private void SendRawData(ChatClientCollection _clientCollection, byte[] _data)
				{					
					foreach (ChatClient _client in _clientCollection)
						SendRawData(_client, _data);
				}
				
				// this function if for debugging purposes only
				private void Log(string strError)
				{
					System.Diagnostics.Debug.Write(strError);
				}			
			}
		}
	}
}
